@@ -32,9 +32,11 @@ import java.nio.ByteBuffer;
 import java.nio.channels.FileChannel;
 import java.nio.channels.WritableByteChannel;
 
+import org.apache.http.annotation.NotThreadSafe;
 import org.apache.http.impl.io.HttpTransportMetricsImpl;
 import org.apache.http.nio.FileContentEncoder;
 import org.apache.http.nio.reactor.SessionOutputBuffer;
+import org.apache.http.util.Args;
 
 /**
  * Content encoder that cuts off after a defined number of bytes. This class
@@ -49,24 +51,49 @@ import org.apache.http.nio.reactor.SessionOutputBuffer;
  *
  * @since 4.0
  */
+@NotThreadSafe
 public class LengthDelimitedEncoder extends AbstractContentEncoder
         implements FileContentEncoder {
 
     private final long contentLength;
+    private final int fragHint;
 
-    private long len;
+    private long remaining;
+
+    /**
+     * @since 4.3
+     *
+     * @param channel underlying channel.
+     * @param buffer  session buffer.
+     * @param metrics transport metrics.
+     * @param contentLength content length.
+     * @param fragementSizeHint fragment size hint defining an minimal size of a fragment
+     *   that should be written out directly to the channel bypassing the session buffer.
+     *   Value <code>0</code> disables fragment buffering.
+     */
+    public LengthDelimitedEncoder(
+            final WritableByteChannel channel,
+            final SessionOutputBuffer buffer,
+            final HttpTransportMetricsImpl metrics,
+            final long contentLength,
+            final int fragementSizeHint) {
+        super(channel, buffer, metrics);
+        Args.notNegative(contentLength, "Content length");
+        this.contentLength = contentLength;
+        this.fragHint = fragementSizeHint > 0 ? fragementSizeHint : 0;
+        this.remaining = contentLength;
+    }
 
     public LengthDelimitedEncoder(
             final WritableByteChannel channel,
             final SessionOutputBuffer buffer,
             final HttpTransportMetricsImpl metrics,
-            long contentLength) {
-        super(channel, buffer, metrics);
-        if (contentLength < 0) {
-            throw new IllegalArgumentException("Content length may not be negative");
-        }
-        this.contentLength = contentLength;
-        this.len = 0;
+            final long contentLength) {
+        this(channel, buffer, metrics, contentLength, 0);
+    }
+
+    private int nextChunk(final ByteBuffer src) {
+        return (int) Math.min(Math.min(this.remaining, Integer.MAX_VALUE), src.remaining());
     }
 
     public int write(final ByteBuffer src) throws IOException {
@@ -74,63 +101,84 @@ public class LengthDelimitedEncoder extends AbstractContentEncoder
             return 0;
         }
         assertNotCompleted();
-        int lenRemaining = (int) (this.contentLength - this.len);
 
-        int bytesWritten;
-        if (src.remaining() > lenRemaining) {
-            int oldLimit = src.limit();
-            int newLimit = oldLimit - (src.remaining() - lenRemaining);
-            src.limit(newLimit);
-            bytesWritten = this.channel.write(src);
-            src.limit(oldLimit);
-        } else {
-            bytesWritten = this.channel.write(src);
+        int total = 0;
+        while (src.hasRemaining() && this.remaining > 0) {
+            if (this.buffer.hasData() || this.fragHint > 0) {
+                final int chunk = nextChunk(src);
+                if (chunk <= this.fragHint) {
+                    final int capacity = this.fragHint - this.buffer.length();
+                    if (capacity > 0) {
+                        final int limit = Math.min(capacity, chunk);
+                        final int bytesWritten = writeToBuffer(src, limit);
+                        this.remaining -= bytesWritten;
+                        total += bytesWritten;
+                    }
+                }
+            }
+            if (this.buffer.hasData()) {
+                final int chunk = nextChunk(src);
+                if (this.buffer.length() >= this.fragHint || chunk > 0) {
+                    final int bytesWritten = flushToChannel();
+                    if (bytesWritten == 0) {
+                        break;
+                    }
+                }
+            }
+            if (!this.buffer.hasData()) {
+                final int chunk = nextChunk(src);
+                if (chunk > this.fragHint) {
+                    final int bytesWritten = writeToChannel(src, chunk);
+                    this.remaining -= bytesWritten;
+                    total += bytesWritten;
+                    if (bytesWritten == 0) {
+                        break;
+                    }
+                }
+            }
         }
-        if (bytesWritten > 0) {
-            this.metrics.incrementBytesTransferred(bytesWritten);
+        if (this.remaining <= 0) {
+            super.complete();
         }
-        this.len += bytesWritten;
-        if (this.len >= this.contentLength) {
-            this.completed = true;
-        }
-        return bytesWritten;
+        return total;
     }
 
     public long transfer(
             final FileChannel src,
-            long position,
-            long count) throws IOException {
+            final long position,
+            final long count) throws IOException {
 
         if (src == null) {
             return 0;
         }
         assertNotCompleted();
-        int lenRemaining = (int) (this.contentLength - this.len);
 
-        long bytesWritten;
-        if (count > lenRemaining) {
-            count = lenRemaining;
+        flushToChannel();
+        if (this.buffer.hasData()) {
+            return 0;
         }
-        bytesWritten = src.transferTo(position, count, this.channel);
+
+        final long chunk = Math.min(this.remaining, count);
+        final long bytesWritten = src.transferTo(position, chunk, this.channel);
         if (bytesWritten > 0) {
             this.metrics.incrementBytesTransferred(bytesWritten);
         }
-        this.len += bytesWritten;
-        if (this.len >= this.contentLength) {
-            this.completed = true;
+        this.remaining -= bytesWritten;
+        if (this.remaining <= 0) {
+            super.complete();
         }
         return bytesWritten;
     }
 
     @Override
     public String toString() {
-        StringBuffer buffer = new StringBuffer();
+        final StringBuilder buffer = new StringBuilder();
         buffer.append("[content length: ");
         buffer.append(this.contentLength);
         buffer.append("; pos: ");
-        buffer.append(this.len);
+        buffer.append(this.contentLength - this.remaining);
         buffer.append("; completed: ");
-        buffer.append(this.completed);
+        buffer.append(isCompleted());
         buffer.append("]");
         return buffer.toString();
     }
