@@ -35,40 +35,29 @@ import java.net.UnknownHostException;
 import java.nio.channels.CancelledKeyException;
 import java.nio.channels.SelectionKey;
 import java.nio.channels.SocketChannel;
-import java.util.Iterator;
 import java.util.Queue;
 import java.util.Set;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.ThreadFactory;
 
+import org.apache.http.annotation.ThreadSafe;
 import org.apache.http.nio.reactor.ConnectingIOReactor;
 import org.apache.http.nio.reactor.IOReactorException;
 import org.apache.http.nio.reactor.IOReactorStatus;
 import org.apache.http.nio.reactor.SessionRequest;
 import org.apache.http.nio.reactor.SessionRequestCallback;
-import org.apache.http.params.HttpConnectionParams;
 import org.apache.http.params.HttpParams;
+import org.apache.http.util.Asserts;
 
 /**
  * Default implementation of {@link ConnectingIOReactor}. This class extends
  * {@link AbstractMultiworkerIOReactor} with capability to connect to remote
  * hosts.
- * <p>
- * The following parameters can be used to customize the behavior of this
- * class:
- * <ul>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#TCP_NODELAY}</li>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#SO_TIMEOUT}</li>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#CONNECTION_TIMEOUT}</li>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#SO_LINGER}</li>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#SO_REUSEADDR}</li>
- *  <li>{@link org.apache.http.nio.params.NIOReactorPNames#SELECT_INTERVAL}</li>
- *  <li>{@link org.apache.http.nio.params.NIOReactorPNames#GRACE_PERIOD}</li>
- *  <li>{@link org.apache.http.nio.params.NIOReactorPNames#INTEREST_OPS_QUEUEING}</li>
- * </ul>
  *
  * @since 4.0
  */
+@SuppressWarnings("deprecation")
+@ThreadSafe // public methods only
 public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
         implements ConnectingIOReactor {
 
@@ -76,19 +65,67 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
 
     private long lastTimeoutCheck;
 
+    /**
+     * Creates an instance of DefaultConnectingIOReactor with the given configuration.
+     *
+     * @param config I/O reactor configuration.
+     * @param threadFactory the factory to create threads.
+     *   Can be <code>null</code>.
+     * @throws IOReactorException in case if a non-recoverable I/O error.
+     *
+     * @since 4.2
+     */
     public DefaultConnectingIOReactor(
-            int workerCount,
-            final ThreadFactory threadFactory,
-            final HttpParams params) throws IOReactorException {
-        super(workerCount, threadFactory, params);
+            final IOReactorConfig config,
+            final ThreadFactory threadFactory) throws IOReactorException {
+        super(config, threadFactory);
         this.requestQueue = new ConcurrentLinkedQueue<SessionRequestImpl>();
         this.lastTimeoutCheck = System.currentTimeMillis();
     }
 
+    /**
+     * Creates an instance of DefaultConnectingIOReactor with the given configuration.
+     *
+     * @param config I/O reactor configuration.
+     *   Can be <code>null</code>.
+     * @throws IOReactorException in case if a non-recoverable I/O error.
+     *
+     * @since 4.2
+     */
+    public DefaultConnectingIOReactor(final IOReactorConfig config) throws IOReactorException {
+        this(config, null);
+    }
+
+    /**
+     * Creates an instance of DefaultConnectingIOReactor with default configuration.
+     *
+     * @throws IOReactorException in case if a non-recoverable I/O error.
+     *
+     * @since 4.2
+     */
+    public DefaultConnectingIOReactor() throws IOReactorException {
+        this(null, null);
+    }
+
+    /**
+     * @deprecated (4.2) use {@link DefaultConnectingIOReactor#DefaultConnectingIOReactor(IOReactorConfig, ThreadFactory)}
+     */
+    @Deprecated
     public DefaultConnectingIOReactor(
-            int workerCount,
+            final int workerCount,
+            final ThreadFactory threadFactory,
             final HttpParams params) throws IOReactorException {
-        this(workerCount, null, params);
+        this(convert(workerCount, params), threadFactory);
+    }
+
+    /**
+     * @deprecated (4.2) use {@link DefaultConnectingIOReactor#DefaultConnectingIOReactor(IOReactorConfig)}
+     */
+    @Deprecated
+    public DefaultConnectingIOReactor(
+            final int workerCount,
+            final HttpParams params) throws IOReactorException {
+        this(convert(workerCount, params), null);
     }
 
     @Override
@@ -100,24 +137,23 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
     }
 
     @Override
-    protected void processEvents(int readyCount) throws IOReactorException {
+    protected void processEvents(final int readyCount) throws IOReactorException {
         processSessionRequests();
 
         if (readyCount > 0) {
-            Set<SelectionKey> selectedKeys = this.selector.selectedKeys();
-            for (Iterator<SelectionKey> it = selectedKeys.iterator(); it.hasNext(); ) {
+            final Set<SelectionKey> selectedKeys = this.selector.selectedKeys();
+            for (final SelectionKey key : selectedKeys) {
 
-                SelectionKey key = it.next();
                 processEvent(key);
 
             }
             selectedKeys.clear();
         }
 
-        long currentTime = System.currentTimeMillis();
+        final long currentTime = System.currentTimeMillis();
         if ((currentTime - this.lastTimeoutCheck) >= this.selectTimeout) {
             this.lastTimeoutCheck = currentTime;
-            Set<SelectionKey> keys = this.selector.keys();
+            final Set<SelectionKey> keys = this.selector.keys();
             processTimeouts(keys);
         }
     }
@@ -127,15 +163,15 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
 
             if (key.isConnectable()) {
 
-                SocketChannel channel = (SocketChannel) key.channel();
+                final SocketChannel channel = (SocketChannel) key.channel();
                 // Get request handle
-                SessionRequestHandle requestHandle = (SessionRequestHandle) key.attachment();
-                SessionRequestImpl sessionRequest = requestHandle.getSessionRequest();
+                final SessionRequestHandle requestHandle = (SessionRequestHandle) key.attachment();
+                final SessionRequestImpl sessionRequest = requestHandle.getSessionRequest();
 
                 // Finish connection process
                 try {
                     channel.finishConnect();
-                } catch (IOException ex) {
+                } catch (final IOException ex) {
                     sessionRequest.failed(ex);
                 }
                 key.cancel();
@@ -143,36 +179,35 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
                     try {
                         try {
                             prepareSocket(channel.socket());
-                        } catch (IOException ex) {
+                        } catch (final IOException ex) {
                             if (this.exceptionHandler == null
                                     || !this.exceptionHandler.handle(ex)) {
                                 throw new IOReactorException(
                                         "Failure initalizing socket", ex);
                             }
                         }
-                        ChannelEntry entry = new ChannelEntry(channel, sessionRequest);
+                        final ChannelEntry entry = new ChannelEntry(channel, sessionRequest);
                         addChannel(entry);
-                    } catch (IOException ex) {
+                    } catch (final IOException ex) {
                         sessionRequest.failed(ex);
                     }
                 }
             }
 
-        } catch (CancelledKeyException ex) {
+        } catch (final CancelledKeyException ex) {
             key.attach(null);
         }
     }
 
     private void processTimeouts(final Set<SelectionKey> keys) {
-        long now = System.currentTimeMillis();
-        for (Iterator<SelectionKey> it = keys.iterator(); it.hasNext();) {
-            SelectionKey key = it.next();
-            Object attachment = key.attachment();
+        final long now = System.currentTimeMillis();
+        for (final SelectionKey key : keys) {
+            final Object attachment = key.attachment();
 
             if (attachment instanceof SessionRequestHandle) {
-                SessionRequestHandle handle = (SessionRequestHandle) key.attachment();
-                SessionRequestImpl sessionRequest = handle.getSessionRequest();
-                int timeout = sessionRequest.getConnectTimeout();
+                final SessionRequestHandle handle = (SessionRequestHandle) key.attachment();
+                final SessionRequestImpl sessionRequest = handle.getSessionRequest();
+                final int timeout = sessionRequest.getConnectTimeout();
                 if (timeout > 0) {
                     if (handle.getRequestTime() + timeout < now) {
                         sessionRequest.timeout();
@@ -188,13 +223,11 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
             final SocketAddress localAddress,
             final Object attachment,
             final SessionRequestCallback callback) {
-
-        if (this.status.compareTo(IOReactorStatus.ACTIVE) > 0) {
-            throw new IllegalStateException("I/O reactor has been shut down");
-        }
-        SessionRequestImpl sessionRequest = new SessionRequestImpl(
+        Asserts.check(this.status.compareTo(IOReactorStatus.ACTIVE) <= 0,
+            "I/O reactor has been shut down");
+        final SessionRequestImpl sessionRequest = new SessionRequestImpl(
                 remoteAddress, localAddress, attachment, callback);
-        sessionRequest.setConnectTimeout(HttpConnectionParams.getConnectionTimeout(this.params));
+        sessionRequest.setConnectTimeout(this.config.getConnectTimeout());
 
         this.requestQueue.add(sessionRequest);
         this.selector.wakeup();
@@ -207,7 +240,7 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
             return;
         }
         if (address instanceof InetSocketAddress) {
-            InetSocketAddress endpoint = (InetSocketAddress) address;
+            final InetSocketAddress endpoint = (InetSocketAddress) address;
             if (endpoint.isUnresolved()) {
                 throw new UnknownHostException(endpoint.getHostName());
             }
@@ -220,10 +253,10 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
             if (request.isCompleted()) {
                 continue;
             }
-            SocketChannel socketChannel;
+            final SocketChannel socketChannel;
             try {
                 socketChannel = SocketChannel.open();
-            } catch (IOException ex) {
+            } catch (final IOException ex) {
                 throw new IOReactorException("Failure opening socket", ex);
             }
             try {
@@ -232,29 +265,29 @@ public class DefaultConnectingIOReactor extends AbstractMultiworkerIOReactor
                 validateAddress(request.getRemoteAddress());
 
                 if (request.getLocalAddress() != null) {
-                    Socket sock = socketChannel.socket();
-                    sock.setReuseAddress(HttpConnectionParams.getSoReuseaddr(this.params));
+                    final Socket sock = socketChannel.socket();
+                    sock.setReuseAddress(this.config.isSoReuseAddress());
                     sock.bind(request.getLocalAddress());
                 }
-                boolean connected = socketChannel.connect(request.getRemoteAddress());
+                final boolean connected = socketChannel.connect(request.getRemoteAddress());
                 if (connected) {
                     prepareSocket(socketChannel.socket());
-                    ChannelEntry entry = new ChannelEntry(socketChannel, request);
+                    final ChannelEntry entry = new ChannelEntry(socketChannel, request);
                     addChannel(entry);
                     return;
                 }
-            } catch (IOException ex) {
+            } catch (final IOException ex) {
                 closeChannel(socketChannel);
                 request.failed(ex);
                 return;
             }
 
-            SessionRequestHandle requestHandle = new SessionRequestHandle(request);
+            final SessionRequestHandle requestHandle = new SessionRequestHandle(request);
             try {
-                SelectionKey key = socketChannel.register(this.selector, SelectionKey.OP_CONNECT,
+                final SelectionKey key = socketChannel.register(this.selector, SelectionKey.OP_CONNECT,
                         requestHandle);
                 request.setKey(key);
-            } catch (IOException ex) {
+            } catch (final IOException ex) {
                 closeChannel(socketChannel);
                 throw new IOReactorException("Failure registering channel " +
                         "with the selector", ex);
