@@ -28,6 +28,7 @@
 package org.apache.http.impl.nio.reactor;
 
 import java.io.IOException;
+import java.net.ServerSocket;
 import java.net.SocketAddress;
 import java.nio.channels.CancelledKeyException;
 import java.nio.channels.SelectionKey;
@@ -41,30 +42,23 @@ import java.util.Set;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.ThreadFactory;
 
+import org.apache.http.annotation.ThreadSafe;
 import org.apache.http.nio.reactor.IOReactorException;
 import org.apache.http.nio.reactor.IOReactorStatus;
 import org.apache.http.nio.reactor.ListenerEndpoint;
 import org.apache.http.nio.reactor.ListeningIOReactor;
 import org.apache.http.params.HttpParams;
+import org.apache.http.util.Asserts;
 
 /**
  * Default implementation of {@link ListeningIOReactor}. This class extends
  * {@link AbstractMultiworkerIOReactor} with capability to listen for incoming
  * connections.
- * <p>
- * The following parameters can be used to customize the behavior of this
- * class:
- * <ul>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#TCP_NODELAY}</li>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#SO_TIMEOUT}</li>
- *  <li>{@link org.apache.http.params.CoreConnectionPNames#SO_LINGER}</li>
- *  <li>{@link org.apache.http.nio.params.NIOReactorPNames#SELECT_INTERVAL}</li>
- *  <li>{@link org.apache.http.nio.params.NIOReactorPNames#GRACE_PERIOD}</li>
- *  <li>{@link org.apache.http.nio.params.NIOReactorPNames#INTEREST_OPS_QUEUEING}</li>
- * </ul>
  *
  * @since 4.0
  */
+@SuppressWarnings("deprecation")
+@ThreadSafe // public methods only
 public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
         implements ListeningIOReactor {
 
@@ -74,22 +68,69 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
 
     private volatile boolean paused;
 
+    /**
+     * Creates an instance of DefaultListeningIOReactor with the given configuration.
+     *
+     * @param config I/O reactor configuration.
+     * @param threadFactory the factory to create threads.
+     *   Can be <code>null</code>.
+     * @throws IOReactorException in case if a non-recoverable I/O error.
+     *
+     * @since 4.2
+     */
     public DefaultListeningIOReactor(
-            int workerCount,
-            final ThreadFactory threadFactory,
-            final HttpParams params) throws IOReactorException {
-        super(workerCount, threadFactory, params);
+            final IOReactorConfig config,
+            final ThreadFactory threadFactory) throws IOReactorException {
+        super(config, threadFactory);
         this.requestQueue = new ConcurrentLinkedQueue<ListenerEndpointImpl>();
         this.endpoints = Collections.synchronizedSet(new HashSet<ListenerEndpointImpl>());
         this.pausedEndpoints = new HashSet<SocketAddress>();
     }
 
-    public DefaultListeningIOReactor(
-            int workerCount,
-            final HttpParams params) throws IOReactorException {
-        this(workerCount, null, params);
+    /**
+     * Creates an instance of DefaultListeningIOReactor with the given configuration.
+     *
+     * @param config I/O reactor configuration.
+     *   Can be <code>null</code>.
+     * @throws IOReactorException in case if a non-recoverable I/O error.
+     *
+     * @since 4.2
+     */
+    public DefaultListeningIOReactor(final IOReactorConfig config) throws IOReactorException {
+        this(config, null);
     }
 
+    /**
+     * Creates an instance of DefaultListeningIOReactor with default configuration.
+     *
+     * @throws IOReactorException in case if a non-recoverable I/O error.
+     *
+     * @since 4.2
+     */
+    public DefaultListeningIOReactor() throws IOReactorException {
+        this(null, null);
+    }
+
+    /**
+     * @deprecated (4.2) use {@link DefaultListeningIOReactor#DefaultListeningIOReactor(IOReactorConfig, ThreadFactory)}
+     */
+    @Deprecated
+    public DefaultListeningIOReactor(
+            final int workerCount,
+            final ThreadFactory threadFactory,
+            final HttpParams params) throws IOReactorException {
+        this(convert(workerCount, params), threadFactory);
+    }
+
+    /**
+     * @deprecated (4.2) use {@link DefaultListeningIOReactor#DefaultListeningIOReactor(IOReactorConfig)}
+     */
+    @Deprecated
+    public DefaultListeningIOReactor(
+            final int workerCount,
+            final HttpParams params) throws IOReactorException {
+        this(convert(workerCount, params), null);
+    }
 
     @Override
     protected void cancelRequests() throws IOReactorException {
@@ -100,16 +141,15 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
     }
 
     @Override
-    protected void processEvents(int readyCount) throws IOReactorException {
+    protected void processEvents(final int readyCount) throws IOReactorException {
         if (!this.paused) {
             processSessionRequests();
         }
 
         if (readyCount > 0) {
-            Set<SelectionKey> selectedKeys = this.selector.selectedKeys();
-            for (Iterator<SelectionKey> it = selectedKeys.iterator(); it.hasNext(); ) {
+            final Set<SelectionKey> selectedKeys = this.selector.selectedKeys();
+            for (final SelectionKey key : selectedKeys) {
 
-                SelectionKey key = it.next();
                 processEvent(key);
 
             }
@@ -123,42 +163,44 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
 
             if (key.isAcceptable()) {
 
-                ServerSocketChannel serverChannel = (ServerSocketChannel) key.channel();
-                SocketChannel socketChannel = null;
-                try {
-                    socketChannel = serverChannel.accept();
-                } catch (IOException ex) {
-                    if (this.exceptionHandler == null ||
-                            !this.exceptionHandler.handle(ex)) {
-                        throw new IOReactorException(
-                                "Failure accepting connection", ex);
+                final ServerSocketChannel serverChannel = (ServerSocketChannel) key.channel();
+                for (;;) {
+                    SocketChannel socketChannel = null;
+                    try {
+                        socketChannel = serverChannel.accept();
+                    } catch (final IOException ex) {
+                        if (this.exceptionHandler == null ||
+                                !this.exceptionHandler.handle(ex)) {
+                            throw new IOReactorException(
+                                    "Failure accepting connection", ex);
+                        }
+                    }
+                    if (socketChannel == null) {
+                        break;
                     }
-                }
-
-                if (socketChannel != null) {
                     try {
                         prepareSocket(socketChannel.socket());
-                    } catch (IOException ex) {
+                    } catch (final IOException ex) {
                         if (this.exceptionHandler == null ||
                                 !this.exceptionHandler.handle(ex)) {
                             throw new IOReactorException(
                                     "Failure initalizing socket", ex);
                         }
                     }
-                    ChannelEntry entry = new ChannelEntry(socketChannel);
+                    final ChannelEntry entry = new ChannelEntry(socketChannel);
                     addChannel(entry);
                 }
             }
 
-        } catch (CancelledKeyException ex) {
-            ListenerEndpoint endpoint = (ListenerEndpoint) key.attachment();
+        } catch (final CancelledKeyException ex) {
+            final ListenerEndpoint endpoint = (ListenerEndpoint) key.attachment();
             this.endpoints.remove(endpoint);
             key.attach(null);
         }
     }
 
     private ListenerEndpointImpl createEndpoint(final SocketAddress address) {
-        ListenerEndpointImpl endpoint = new ListenerEndpointImpl(
+        return new ListenerEndpointImpl(
                 address,
                 new ListenerEndpointClosedCallback() {
 
@@ -167,14 +209,12 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
                     }
 
                 });
-        return endpoint;
     }
 
     public ListenerEndpoint listen(final SocketAddress address) {
-        if (this.status.compareTo(IOReactorStatus.ACTIVE) > 0) {
-            throw new IllegalStateException("I/O reactor has been shut down");
-        }
-        ListenerEndpointImpl request = createEndpoint(address);
+        Asserts.check(this.status.compareTo(IOReactorStatus.ACTIVE) <= 0,
+                "I/O reactor has been shut down");
+        final ListenerEndpointImpl request = createEndpoint(address);
         this.requestQueue.add(request);
         this.selector.wakeup();
         return request;
@@ -183,17 +223,19 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
     private void processSessionRequests() throws IOReactorException {
         ListenerEndpointImpl request;
         while ((request = this.requestQueue.poll()) != null) {
-            SocketAddress address = request.getAddress();
-            ServerSocketChannel serverChannel;
+            final SocketAddress address = request.getAddress();
+            final ServerSocketChannel serverChannel;
             try {
                 serverChannel = ServerSocketChannel.open();
-            } catch (IOException ex) {
+            } catch (final IOException ex) {
                 throw new IOReactorException("Failure opening server socket", ex);
             }
             try {
+                final ServerSocket socket = serverChannel.socket();
+                socket.setReuseAddress(this.config.isSoReuseAddress());
                 serverChannel.configureBlocking(false);
-                serverChannel.socket().bind(address);
-            } catch (IOException ex) {
+                socket.bind(address);
+            } catch (final IOException ex) {
                 closeChannel(serverChannel);
                 request.failed(ex);
                 if (this.exceptionHandler == null || !this.exceptionHandler.handle(ex)) {
@@ -204,10 +246,10 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
                 }
             }
             try {
-                SelectionKey key = serverChannel.register(this.selector, SelectionKey.OP_ACCEPT);
+                final SelectionKey key = serverChannel.register(this.selector, SelectionKey.OP_ACCEPT);
                 key.attach(request);
                 request.setKey(key);
-            } catch (IOException ex) {
+            } catch (final IOException ex) {
                 closeChannel(serverChannel);
                 throw new IOReactorException("Failure registering channel " +
                         "with the selector", ex);
@@ -219,11 +261,11 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
     }
 
     public Set<ListenerEndpoint> getEndpoints() {
-        Set<ListenerEndpoint> set = new HashSet<ListenerEndpoint>();
+        final Set<ListenerEndpoint> set = new HashSet<ListenerEndpoint>();
         synchronized (this.endpoints) {
-            Iterator<ListenerEndpointImpl> it = this.endpoints.iterator();
+            final Iterator<ListenerEndpointImpl> it = this.endpoints.iterator();
             while (it.hasNext()) {
-                ListenerEndpoint endpoint = it.next();
+                final ListenerEndpoint endpoint = it.next();
                 if (!endpoint.isClosed()) {
                     set.add(endpoint);
                 } else {
@@ -240,9 +282,7 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
         }
         this.paused = true;
         synchronized (this.endpoints) {
-            Iterator<ListenerEndpointImpl> it = this.endpoints.iterator();
-            while (it.hasNext()) {
-                ListenerEndpoint endpoint = it.next();
+            for (final ListenerEndpointImpl endpoint : this.endpoints) {
                 if (!endpoint.isClosed()) {
                     endpoint.close();
                     this.pausedEndpoints.add(endpoint.getAddress());
@@ -257,8 +297,8 @@ public class DefaultListeningIOReactor extends AbstractMultiworkerIOReactor
             return;
         }
         this.paused = false;
-        for (SocketAddress address: this.pausedEndpoints) {
-            ListenerEndpointImpl request = createEndpoint(address);
+        for (final SocketAddress address: this.pausedEndpoints) {
+            final ListenerEndpointImpl request = createEndpoint(address);
             this.requestQueue.add(request);
         }
         this.pausedEndpoints.clear();
