@@ -32,11 +32,14 @@ import java.io.InputStream;
 import java.io.OutputStream;
 
 import org.apache.http.HttpEntity;
+import org.apache.http.annotation.NotThreadSafe;
 import org.apache.http.entity.HttpEntityWrapper;
 import org.apache.http.nio.ContentDecoder;
 import org.apache.http.nio.IOControl;
 import org.apache.http.nio.util.ByteBufferAllocator;
 import org.apache.http.nio.util.SimpleInputBuffer;
+import org.apache.http.util.Args;
+import org.apache.http.util.Asserts;
 
 /**
  * A {@link ConsumingNHttpEntity} that consumes content into a buffer. The
@@ -45,7 +48,13 @@ import org.apache.http.nio.util.SimpleInputBuffer;
  * {@link HttpEntity#writeTo(OutputStream)}.
  *
  * @since 4.0
+ *
+ * @deprecated use (4.2)
+ *  {@link org.apache.http.nio.protocol.BasicAsyncRequestProducer}
+ *  or {@link org.apache.http.nio.protocol.BasicAsyncResponseProducer}
  */
+@NotThreadSafe
+@Deprecated
 public class BufferingNHttpEntity extends HttpEntityWrapper implements
         ConsumingNHttpEntity {
 
@@ -75,10 +84,6 @@ public class BufferingNHttpEntity extends HttpEntityWrapper implements
         this.finished = true;
     }
 
-    @Override
-    public void consumeContent() throws IOException {
-    }
-
     /**
      * Obtains entity's content as {@link InputStream}.
      *
@@ -87,12 +92,8 @@ public class BufferingNHttpEntity extends HttpEntityWrapper implements
      */
     @Override
     public InputStream getContent() throws IOException {
-        if (!this.finished) {
-            throw new IllegalStateException("Entity content has not been fully received");
-        }
-        if (this.consumed) {
-            throw new IllegalStateException("Entity content has been consumed");
-        }
+        Asserts.check(this.finished, "Entity content has not been fully received");
+        Asserts.check(!this.consumed, "Entity content has been consumed");
         this.consumed = true;
         return new ContentInputStream(this.buffer);
     }
@@ -109,11 +110,9 @@ public class BufferingNHttpEntity extends HttpEntityWrapper implements
 
     @Override
     public void writeTo(final OutputStream outstream) throws IOException {
-        if (outstream == null) {
-            throw new IllegalArgumentException("Output stream may not be null");
-        }
-        InputStream instream = getContent();
-        byte[] buffer = new byte[BUFFER_SIZE];
+        Args.notNull(outstream, "Output stream");
+        final InputStream instream = getContent();
+        final byte[] buffer = new byte[BUFFER_SIZE];
         int l;
         // consume until EOF
         while ((l = instream.read(buffer)) != -1) {
