@@ -35,9 +35,11 @@ import java.util.LinkedList;
 import java.util.ListIterator;
 import java.util.Map;
 import java.util.Set;
+import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.Future;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.TimeoutException;
+import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.locks.Lock;
 import java.util.concurrent.locks.ReentrantLock;
 
@@ -51,7 +53,10 @@ import org.apache.http.nio.reactor.SessionRequestCallback;
 import org.apache.http.pool.ConnPool;
 import org.apache.http.pool.ConnPoolControl;
 import org.apache.http.pool.PoolEntry;
+import org.apache.http.pool.PoolEntryCallback;
 import org.apache.http.pool.PoolStats;
+import org.apache.http.util.Args;
+import org.apache.http.util.Asserts;
 
 /**
  * Abstract non-blocking connection pool.
@@ -68,39 +73,49 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
 
     private final ConnectingIOReactor ioreactor;
     private final NIOConnFactory<T, C> connFactory;
+    private final SocketAddressResolver<T> addressResolver;
     private final SessionRequestCallback sessionRequestCallback;
     private final Map<T, RouteSpecificPool<T, C, E>> routeToPool;
     private final LinkedList<LeaseRequest<T, C, E>> leasingRequests;
     private final Set<SessionRequest> pending;
     private final Set<E> leased;
     private final LinkedList<E> available;
+    private final ConcurrentLinkedQueue<LeaseRequest<T, C, E>> completedRequests;
     private final Map<T, Integer> maxPerRoute;
     private final Lock lock;
+    private final AtomicBoolean isShutDown;
 
-    private volatile boolean isShutDown;
     private volatile int defaultMaxPerRoute;
     private volatile int maxTotal;
 
+    /**
+     * @deprecated use {@link AbstractNIOConnPool#AbstractNIOConnPool(ConnectingIOReactor,
+     *   NIOConnFactory, SocketAddressResolver, int, int)}
+     */
+    @Deprecated
     public AbstractNIOConnPool(
             final ConnectingIOReactor ioreactor,
             final NIOConnFactory<T, C> connFactory,
-            int defaultMaxPerRoute,
-            int maxTotal) {
+            final int defaultMaxPerRoute,
+            final int maxTotal) {
         super();
-        if (ioreactor == null) {
-            throw new IllegalArgumentException("I/O reactor may not be null");
-        }
-        if (connFactory == null) {
-            throw new IllegalArgumentException("Connection factory may not null");
-        }
-        if (defaultMaxPerRoute <= 0) {
-            throw new IllegalArgumentException("Max per route value may not be negative or zero");
-        }
-        if (maxTotal <= 0) {
-            throw new IllegalArgumentException("Max total value may not be negative or zero");
-        }
+        Args.notNull(ioreactor, "I/O reactor");
+        Args.notNull(connFactory, "Connection factory");
+        Args.positive(defaultMaxPerRoute, "Max per route value");
+        Args.positive(maxTotal, "Max total value");
         this.ioreactor = ioreactor;
         this.connFactory = connFactory;
+        this.addressResolver = new SocketAddressResolver<T>() {
+
+            public SocketAddress resolveLocalAddress(final T route) throws IOException {
+                return AbstractNIOConnPool.this.resolveLocalAddress(route);
+            }
+
+            public SocketAddress resolveRemoteAddress(final T route) throws IOException {
+                return AbstractNIOConnPool.this.resolveRemoteAddress(route);
+            }
+
+        };
         this.sessionRequestCallback = new InternalSessionRequestCallback();
         this.routeToPool = new HashMap<T, RouteSpecificPool<T, C, E>>();
         this.leasingRequests = new LinkedList<LeaseRequest<T, C, E>>();
@@ -108,48 +123,105 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
         this.leased = new HashSet<E>();
         this.available = new LinkedList<E>();
         this.maxPerRoute = new HashMap<T, Integer>();
+        this.completedRequests = new ConcurrentLinkedQueue<LeaseRequest<T, C, E>>();
         this.lock = new ReentrantLock();
+        this.isShutDown = new AtomicBoolean(false);
         this.defaultMaxPerRoute = defaultMaxPerRoute;
         this.maxTotal = maxTotal;
     }
 
-    protected abstract SocketAddress resolveRemoteAddress(T route);
+    /**
+     * @since 4.3
+     */
+    public AbstractNIOConnPool(
+            final ConnectingIOReactor ioreactor,
+            final NIOConnFactory<T, C> connFactory,
+            final SocketAddressResolver<T> addressResolver,
+            final int defaultMaxPerRoute,
+            final int maxTotal) {
+        super();
+        Args.notNull(ioreactor, "I/O reactor");
+        Args.notNull(connFactory, "Connection factory");
+        Args.notNull(addressResolver, "Address resolver");
+        Args.positive(defaultMaxPerRoute, "Max per route value");
+        Args.positive(maxTotal, "Max total value");
+        this.ioreactor = ioreactor;
+        this.connFactory = connFactory;
+        this.addressResolver = addressResolver;
+        this.sessionRequestCallback = new InternalSessionRequestCallback();
+        this.routeToPool = new HashMap<T, RouteSpecificPool<T, C, E>>();
+        this.leasingRequests = new LinkedList<LeaseRequest<T, C, E>>();
+        this.pending = new HashSet<SessionRequest>();
+        this.leased = new HashSet<E>();
+        this.available = new LinkedList<E>();
+        this.completedRequests = new ConcurrentLinkedQueue<LeaseRequest<T, C, E>>();
+        this.maxPerRoute = new HashMap<T, Integer>();
+        this.lock = new ReentrantLock();
+        this.isShutDown = new AtomicBoolean(false);
+        this.defaultMaxPerRoute = defaultMaxPerRoute;
+        this.maxTotal = maxTotal;
+    }
 
-    protected abstract SocketAddress resolveLocalAddress(T route);
+    /**
+     * @deprecated (4.3) use {@link SocketAddressResolver}
+     */
+    @Deprecated
+    protected SocketAddress resolveRemoteAddress(final T route) {
+        return null;
+    }
+
+    /**
+     * @deprecated (4.3) use {@link SocketAddressResolver}
+     */
+    @Deprecated
+    protected SocketAddress resolveLocalAddress(final T route) {
+        return null;
+    }
 
     protected abstract E createEntry(T route, C conn);
 
-    public boolean isShutdown() {
-        return this.isShutDown;
+    /**
+     * @since 4.3
+     */
+    protected void onLease(final E entry) {
     }
 
-    public void shutdown(long waitMs) throws IOException {
-        if (this.isShutDown) {
-            return ;
-        }
-        this.isShutDown = true;
-        this.lock.lock();
-        try {
-            for (SessionRequest sessionRequest: this.pending) {
-                sessionRequest.cancel();
+    /**
+     * @since 4.3
+     */
+    protected void onRelease(final E entry) {
+    }
+
+    public boolean isShutdown() {
+        return this.isShutDown.get();
+    }
+
+    public void shutdown(final long waitMs) throws IOException {
+        if (this.isShutDown.compareAndSet(false, true)) {
+            fireCallbacks();
+            this.lock.lock();
+            try {
+                for (final SessionRequest sessionRequest: this.pending) {
+                    sessionRequest.cancel();
+                }
+                for (final E entry: this.available) {
+                    entry.close();
+                }
+                for (final E entry: this.leased) {
+                    entry.close();
+                }
+                for (final RouteSpecificPool<T, C, E> pool: this.routeToPool.values()) {
+                    pool.shutdown();
+                }
+                this.routeToPool.clear();
+                this.leased.clear();
+                this.pending.clear();
+                this.available.clear();
+                this.leasingRequests.clear();
+                this.ioreactor.shutdown(waitMs);
+            } finally {
+                this.lock.unlock();
             }
-            for (E entry: this.available) {
-                entry.close();
-            }
-            for (E entry: this.leased) {
-                entry.close();
-            }
-            for (RouteSpecificPool<T, C, E> pool: this.routeToPool.values()) {
-                pool.shutdown();
-            }
-            this.routeToPool.clear();
-            this.leased.clear();
-            this.pending.clear();
-            this.available.clear();
-            this.leasingRequests.clear();
-            this.ioreactor.shutdown(waitMs);
-        } finally {
-            this.lock.unlock();
         }
     }
 
@@ -173,27 +245,36 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
             final T route, final Object state,
             final long connectTimeout, final TimeUnit tunit,
             final FutureCallback<E> callback) {
-        if (route == null) {
-            throw new IllegalArgumentException("Route may not be null");
-        }
-        if (tunit == null) {
-            throw new IllegalArgumentException("Time unit may not be null.");
-        }
-        if (this.isShutDown) {
-            throw new IllegalStateException("Session pool has been shut down");
-        }
+        return this.lease(route, state, connectTimeout, connectTimeout, tunit, callback);
+    }
+
+    /**
+     * @since 4.3
+     */
+    public Future<E> lease(
+            final T route, final Object state,
+            final long connectTimeout, final long leaseTimeout, final TimeUnit tunit,
+            final FutureCallback<E> callback) {
+        Args.notNull(route, "Route");
+        Args.notNull(tunit, "Time unit");
+        Asserts.check(!this.isShutDown.get(), "Connection pool shut down");
+        final BasicFuture<E> future = new BasicFuture<E>(callback);
         this.lock.lock();
         try {
-            long timeout = connectTimeout > 0 ? tunit.toMillis(connectTimeout) : 0;
-            BasicFuture<E> future = new BasicFuture<E>(callback);
-            LeaseRequest<T, C, E> request = new LeaseRequest<T, C, E>(route, state, timeout, future);
-            this.leasingRequests.add(request);
-
-            processPendingRequests();
-            return future;
+            final long timeout = connectTimeout > 0 ? tunit.toMillis(connectTimeout) : 0;
+            final LeaseRequest<T, C, E> request = new LeaseRequest<T, C, E>(route, state, timeout, leaseTimeout, future);
+            final boolean completed = processPendingRequest(request);
+            if (!request.isDone() && !completed) {
+                this.leasingRequests.add(request);
+            }
+            if (request.isDone()) {
+                this.completedRequests.add(request);
+            }
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
+        return future;
     }
 
     public Future<E> lease(final T route, final Object state, final FutureCallback<E> callback) {
@@ -204,112 +285,164 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
         return lease(route, state, -1, TimeUnit.MICROSECONDS, null);
     }
 
-    public void release(final E entry, boolean reusable) {
+    public void release(final E entry, final boolean reusable) {
         if (entry == null) {
             return;
         }
-        if (this.isShutDown) {
+        if (this.isShutDown.get()) {
             return;
         }
         this.lock.lock();
         try {
             if (this.leased.remove(entry)) {
-                RouteSpecificPool<T, C, E> pool = getPool(entry.getRoute());
+                final RouteSpecificPool<T, C, E> pool = getPool(entry.getRoute());
                 pool.free(entry, reusable);
                 if (reusable) {
                     this.available.addFirst(entry);
+                    onRelease(entry);
                 } else {
                     entry.close();
                 }
-                processPendingRequests();
+                processNextPendingRequest();
             }
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
     }
 
     private void processPendingRequests() {
-        ListIterator<LeaseRequest<T, C, E>> it = this.leasingRequests.listIterator();
+        final ListIterator<LeaseRequest<T, C, E>> it = this.leasingRequests.listIterator();
         while (it.hasNext()) {
-            LeaseRequest<T, C, E> request = it.next();
-
-            T route = request.getRoute();
-            Object state = request.getState();
-            long deadline = request.getDeadline();
-            BasicFuture<E> future = request.getFuture();
-
-            long now = System.currentTimeMillis();
-            if (now > deadline) {
+            final LeaseRequest<T, C, E> request = it.next();
+            final boolean completed = processPendingRequest(request);
+            if (request.isDone() || completed) {
                 it.remove();
-                future.failed(new TimeoutException());
-                continue;
             }
+            if (request.isDone()) {
+                this.completedRequests.add(request);
+            }
+        }
+    }
 
-            RouteSpecificPool<T, C, E> pool = getPool(route);
-            E entry = null;
-            for (;;) {
-                entry = pool.getFree(state);
-                if (entry == null) {
-                    break;
-                }
-                if (entry.isClosed() || entry.isExpired(System.currentTimeMillis())) {
-                    entry.close();
-                    this.available.remove(entry);
-                    pool.free(entry, false);
-                } else {
-                    break;
-                }
-            }
-            if (entry != null) {
+    private void processNextPendingRequest() {
+        final ListIterator<LeaseRequest<T, C, E>> it = this.leasingRequests.listIterator();
+        while (it.hasNext()) {
+            final LeaseRequest<T, C, E> request = it.next();
+            final boolean completed = processPendingRequest(request);
+            if (request.isDone() || completed) {
                 it.remove();
+            }
+            if (request.isDone()) {
+                this.completedRequests.add(request);
+            }
+            if (completed) {
+                return;
+            }
+        }
+    }
+
+    private boolean processPendingRequest(final LeaseRequest<T, C, E> request) {
+        final T route = request.getRoute();
+        final Object state = request.getState();
+        final long deadline = request.getDeadline();
+
+        final long now = System.currentTimeMillis();
+        if (now > deadline) {
+            request.failed(new TimeoutException());
+            return false;
+        }
+
+        final RouteSpecificPool<T, C, E> pool = getPool(route);
+        E entry;
+        for (;;) {
+            entry = pool.getFree(state);
+            if (entry == null) {
+                break;
+            }
+            if (entry.isClosed() || entry.isExpired(System.currentTimeMillis())) {
+                entry.close();
                 this.available.remove(entry);
-                this.leased.add(entry);
-                future.completed(entry);
-                continue;
+                pool.free(entry, false);
+            } else {
+                break;
             }
+        }
+        if (entry != null) {
+            this.available.remove(entry);
+            this.leased.add(entry);
+            request.completed(entry);
+            onLease(entry);
+            return true;
+        }
 
-            // New connection is needed
-            int maxPerRoute = getMax(route);
-            // Shrink the pool prior to allocating a new connection
-            int excess = Math.max(0, pool.getAllocatedCount() + 1 - maxPerRoute);
-            if (excess > 0) {
-                for (int i = 0; i < excess; i++) {
-                    E lastUsed = pool.getLastUsed();
-                    if (lastUsed == null) {
-                        break;
-                    }
+        // New connection is needed
+        final int maxPerRoute = getMax(route);
+        // Shrink the pool prior to allocating a new connection
+        final int excess = Math.max(0, pool.getAllocatedCount() + 1 - maxPerRoute);
+        if (excess > 0) {
+            for (int i = 0; i < excess; i++) {
+                final E lastUsed = pool.getLastUsed();
+                if (lastUsed == null) {
+                    break;
+                }
+                lastUsed.close();
+                this.available.remove(lastUsed);
+                pool.remove(lastUsed);
+            }
+        }
+
+        if (pool.getAllocatedCount() < maxPerRoute) {
+            final int totalUsed = this.pending.size() + this.leased.size();
+            final int freeCapacity = Math.max(this.maxTotal - totalUsed, 0);
+            if (freeCapacity == 0) {
+                return false;
+            }
+            final int totalAvailable = this.available.size();
+            if (totalAvailable > freeCapacity - 1) {
+                if (!this.available.isEmpty()) {
+                    final E lastUsed = this.available.removeLast();
                     lastUsed.close();
-                    this.available.remove(lastUsed);
-                    pool.remove(lastUsed);
+                    final RouteSpecificPool<T, C, E> otherpool = getPool(lastUsed.getRoute());
+                    otherpool.remove(lastUsed);
                 }
             }
 
-            if (pool.getAllocatedCount() < maxPerRoute) {
-                int totalUsed = this.pending.size() + this.leased.size();
-                int freeCapacity = Math.max(this.maxTotal - totalUsed, 0);
-                if (freeCapacity == 0) {
-                    continue;
-                }
-                int totalAvailable = this.available.size();
-                if (totalAvailable > freeCapacity - 1) {
-                    if (!this.available.isEmpty()) {
-                        E lastUsed = this.available.removeLast();
-                        lastUsed.close();
-                        RouteSpecificPool<T, C, E> otherpool = getPool(lastUsed.getRoute());
-                        otherpool.remove(lastUsed);
-                    }
-                }
-                it.remove();
-                SessionRequest sessionRequest = this.ioreactor.connect(
-                        resolveRemoteAddress(route),
-                        resolveLocalAddress(route),
-                        route,
-                        this.sessionRequestCallback);
-                int timout = request.getConnectTimeout() < Integer.MAX_VALUE ?
-                        (int) request.getConnectTimeout() : Integer.MAX_VALUE;
-                sessionRequest.setConnectTimeout(timout);
-                this.pending.add(sessionRequest);
-                pool.addPending(sessionRequest, future);
+            final SocketAddress localAddress;
+            final SocketAddress remoteAddress;
+            try {
+                remoteAddress = this.addressResolver.resolveRemoteAddress(route);
+                localAddress = this.addressResolver.resolveLocalAddress(route);
+            } catch (final IOException ex) {
+                request.failed(ex);
+                return false;
+            }
+
+            final SessionRequest sessionRequest = this.ioreactor.connect(
+                    remoteAddress, localAddress, route, this.sessionRequestCallback);
+            final int timout = request.getConnectTimeout() < Integer.MAX_VALUE ?
+                    (int) request.getConnectTimeout() : Integer.MAX_VALUE;
+            sessionRequest.setConnectTimeout(timout);
+            this.pending.add(sessionRequest);
+            pool.addPending(sessionRequest, request.getFuture());
+            return true;
+        } else {
+            return false;
+        }
+    }
+
+    private void fireCallbacks() {
+        LeaseRequest<T, C, E> request;
+        while ((request = this.completedRequests.poll()) != null) {
+            final BasicFuture<E> future = request.getFuture();
+            final Exception ex = request.getException();
+            final E result = request.getResult();
+            if (ex != null) {
+                future.failed(ex);
+            } else if (result != null) {
+                future.completed(result);
+            } else {
+                future.cancel();
             }
         }
     }
@@ -317,100 +450,109 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
     public void validatePendingRequests() {
         this.lock.lock();
         try {
-            long now = System.currentTimeMillis();
-            ListIterator<LeaseRequest<T, C, E>> it = this.leasingRequests.listIterator();
+            final long now = System.currentTimeMillis();
+            final ListIterator<LeaseRequest<T, C, E>> it = this.leasingRequests.listIterator();
             while (it.hasNext()) {
-                LeaseRequest<T, C, E> request = it.next();
-                long deadline = request.getDeadline();
+                final LeaseRequest<T, C, E> request = it.next();
+                final long deadline = request.getDeadline();
                 if (now > deadline) {
                     it.remove();
-                    BasicFuture<E> future = request.getFuture();
-                    future.failed(new TimeoutException());
+                    request.failed(new TimeoutException());
+                    this.completedRequests.add(request);
                 }
             }
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
     }
 
     protected void requestCompleted(final SessionRequest request) {
-        if (this.isShutDown) {
+        if (this.isShutDown.get()) {
             return;
         }
         @SuppressWarnings("unchecked")
+        final
         T route = (T) request.getAttachment();
         this.lock.lock();
         try {
             this.pending.remove(request);
-            RouteSpecificPool<T, C, E> pool = getPool(route);
-            IOSession session = request.getSession();
+            final RouteSpecificPool<T, C, E> pool = getPool(route);
+            final IOSession session = request.getSession();
             try {
-                C conn = this.connFactory.create(route, session);
-                E entry = pool.createEntry(request, conn);
+                final C conn = this.connFactory.create(route, session);
+                final E entry = pool.createEntry(request, conn);
                 this.leased.add(entry);
                 pool.completed(request, entry);
-
-            } catch (IOException ex) {
+                onLease(entry);
+            } catch (final IOException ex) {
                 pool.failed(request, ex);
             }
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
     }
 
     protected void requestCancelled(final SessionRequest request) {
-        if (this.isShutDown) {
+        if (this.isShutDown.get()) {
             return;
         }
         @SuppressWarnings("unchecked")
+        final
         T route = (T) request.getAttachment();
         this.lock.lock();
         try {
             this.pending.remove(request);
-            RouteSpecificPool<T, C, E> pool = getPool(route);
+            final RouteSpecificPool<T, C, E> pool = getPool(route);
             pool.cancelled(request);
-            processPendingRequests();
+            processNextPendingRequest();
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
     }
 
     protected void requestFailed(final SessionRequest request) {
-        if (this.isShutDown) {
+        if (this.isShutDown.get()) {
             return;
         }
         @SuppressWarnings("unchecked")
+        final
         T route = (T) request.getAttachment();
         this.lock.lock();
         try {
             this.pending.remove(request);
-            RouteSpecificPool<T, C, E> pool = getPool(route);
+            final RouteSpecificPool<T, C, E> pool = getPool(route);
             pool.failed(request, request.getException());
-            processPendingRequests();
+            processNextPendingRequest();
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
     }
 
     protected void requestTimeout(final SessionRequest request) {
-        if (this.isShutDown) {
+        if (this.isShutDown.get()) {
             return;
         }
         @SuppressWarnings("unchecked")
+        final
         T route = (T) request.getAttachment();
         this.lock.lock();
         try {
             this.pending.remove(request);
-            RouteSpecificPool<T, C, E> pool = getPool(route);
+            final RouteSpecificPool<T, C, E> pool = getPool(route);
             pool.timeout(request);
-            processPendingRequests();
+            processNextPendingRequest();
         } finally {
             this.lock.unlock();
         }
+        fireCallbacks();
     }
 
     private int getMax(final T route) {
-        Integer v = this.maxPerRoute.get(route);
+        final Integer v = this.maxPerRoute.get(route);
         if (v != null) {
             return v.intValue();
         } else {
@@ -418,10 +560,8 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
         }
     }
 
-    public void setMaxTotal(int max) {
-        if (max <= 0) {
-            throw new IllegalArgumentException("Max value may not be negative or zero");
-        }
+    public void setMaxTotal(final int max) {
+        Args.positive(max, "Max value");
         this.lock.lock();
         try {
             this.maxTotal = max;
@@ -439,10 +579,8 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
         }
     }
 
-    public void setDefaultMaxPerRoute(int max) {
-        if (max <= 0) {
-            throw new IllegalArgumentException("Max value may not be negative or zero");
-        }
+    public void setDefaultMaxPerRoute(final int max) {
+        Args.positive(max, "Max value");
         this.lock.lock();
         try {
             this.defaultMaxPerRoute = max;
@@ -460,25 +598,19 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
         }
     }
 
-    public void setMaxPerRoute(final T route, int max) {
-        if (route == null) {
-            throw new IllegalArgumentException("Route may not be null");
-        }
-        if (max <= 0) {
-            throw new IllegalArgumentException("Max value may not be negative or zero");
-        }
+    public void setMaxPerRoute(final T route, final int max) {
+        Args.notNull(route, "Route");
+        Args.positive(max, "Max value");
         this.lock.lock();
         try {
-            this.maxPerRoute.put(route, max);
+            this.maxPerRoute.put(route, Integer.valueOf(max));
         } finally {
             this.lock.unlock();
         }
     }
 
-    public int getMaxPerRoute(T route) {
-        if (route == null) {
-            throw new IllegalArgumentException("Route may not be null");
-        }
+    public int getMaxPerRoute(final T route) {
+        Args.notNull(route, "Route");
         this.lock.lock();
         try {
             return getMax(route);
@@ -501,12 +633,10 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
     }
 
     public PoolStats getStats(final T route) {
-        if (route == null) {
-            throw new IllegalArgumentException("Route may not be null");
-        }
+        Args.notNull(route, "Route");
         this.lock.lock();
         try {
-            RouteSpecificPool<T, C, E> pool = getPool(route);
+            final RouteSpecificPool<T, C, E> pool = getPool(route);
             return new PoolStats(
                     pool.getLeasedCount(),
                     pool.getPendingCount(),
@@ -517,56 +647,81 @@ public abstract class AbstractNIOConnPool<T, C, E extends PoolEntry<T, C>>
         }
     }
 
-    public void closeIdle(long idletime, final TimeUnit tunit) {
-        if (tunit == null) {
-            throw new IllegalArgumentException("Time unit must not be null.");
+    /**
+     * Enumerates all available connections.
+     *
+     * @since 4.3
+     */
+    protected void enumAvailable(final PoolEntryCallback<T, C> callback) {
+        this.lock.lock();
+        try {
+            enumEntries(this.available.iterator(), callback);
+        } finally {
+            this.lock.unlock();
         }
+    }
+
+    /**
+     * Enumerates all leased connections.
+     *
+     * @since 4.3
+     */
+    protected void enumLeased(final PoolEntryCallback<T, C> callback) {
+        this.lock.lock();
+        try {
+            enumEntries(this.leased.iterator(), callback);
+        } finally {
+            this.lock.unlock();
+        }
+    }
+
+    protected void enumEntries(final Iterator<E> it, final PoolEntryCallback<T, C> callback) {
+        while (it.hasNext()) {
+            final E entry = it.next();
+            callback.process(entry);
+            if (entry.isClosed()) {
+                final RouteSpecificPool<T, C, E> pool = getPool(entry.getRoute());
+                pool.remove(entry);
+                it.remove();
+            }
+        }
+        processPendingRequests();
+    }
+
+    public void closeIdle(final long idletime, final TimeUnit tunit) {
+        Args.notNull(tunit, "Time unit");
         long time = tunit.toMillis(idletime);
         if (time < 0) {
             time = 0;
         }
-        long deadline = System.currentTimeMillis() - time;
-        this.lock.lock();
-        try {
-            Iterator<E> it = this.available.iterator();
-            while (it.hasNext()) {
-                E entry = it.next();
+        final long deadline = System.currentTimeMillis() - time;
+        enumAvailable(new PoolEntryCallback<T, C>() {
+
+            public void process(final PoolEntry<T, C> entry) {
                 if (entry.getUpdated() <= deadline) {
                     entry.close();
-                    RouteSpecificPool<T, C, E> pool = getPool(entry.getRoute());
-                    pool.remove(entry);
-                    it.remove();
                 }
             }
-            processPendingRequests();
-        } finally {
-            this.lock.unlock();
-        }
+
+        });
     }
 
     public void closeExpired() {
-        long now = System.currentTimeMillis();
-        this.lock.lock();
-        try {
-            Iterator<E> it = this.available.iterator();
-            while (it.hasNext()) {
-                E entry = it.next();
+        final long now = System.currentTimeMillis();
+        enumAvailable(new PoolEntryCallback<T, C>() {
+
+            public void process(final PoolEntry<T, C> entry) {
                 if (entry.isExpired(now)) {
                     entry.close();
-                    RouteSpecificPool<T, C, E> pool = getPool(entry.getRoute());
-                    pool.remove(entry);
-                    it.remove();
                 }
             }
-            processPendingRequests();
-        } finally {
-            this.lock.unlock();
-        }
+
+        });
     }
 
     @Override
     public String toString() {
-        StringBuilder buffer = new StringBuilder();
+        final StringBuilder buffer = new StringBuilder();
         buffer.append("[leased: ");
         buffer.append(this.leased);
         buffer.append("][available: ");
